"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

require("core-js/modules/es.promise.js");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.handleErrorAndExit = handleErrorAndExit;
exports.handleUnhandledRejection = handleUnhandledRejection;
exports.logErrAndExit = logErrAndExit;
exports.CommandRunner = void 0;

function _ink() {
  const data = require("ink");

  _ink = function () {
    return data;
  };

  return data;
}

function _serializeError() {
  const data = require("serialize-error");

  _serializeError = function () {
    return data;
  };

  return data;
}

function _consumer() {
  const data = require("../api/consumer");

  _consumer = function () {
    return data;
  };

  return data;
}

function _logger() {
  const data = _interopRequireDefault(require("../logger/logger"));

  _logger = function () {
    return data;
  };

  return data;
}

function _utils() {
  const data = require("../utils");

  _utils = function () {
    return data;
  };

  return data;
}

function _defaultErrorHandler() {
  const data = _interopRequireDefault(require("./default-error-handler"));

  _defaultErrorHandler = function () {
    return data;
  };

  return data;
}

function _loader() {
  const data = _interopRequireDefault(require("./loader"));

  _loader = function () {
    return data;
  };

  return data;
}

class CommandRunner {
  constructor(command, args, flags) {
    this.command = command;
    this.args = args;
    this.flags = flags;
  }
  /**
   * run command using one of the handler, "json"/"report"/"render". once done, exit the process.
   */


  async runCommand() {
    try {
      await this.runMigrateIfNeeded();
      this.determineConsoleWritingDuringCommand();

      if (this.flags.json) {
        return await this.runJsonHandler();
      }

      if (this.shouldRunRender()) {
        return await this.runRenderHandler();
      }

      if (this.command.report) {
        return await this.runReportHandler();
      }
    } catch (err) {
      return handleErrorAndExit(err, this.command.name, this.command.internal);
    }

    throw new Error(`command "${this.command.name}" doesn't implement "render" nor "report" methods`);
  }
  /**
   * when both "render" and "report" were implemented, check whether it's a terminal.
   * if it's a terminal, use "render", if not, use "report" because "report" is just a string
   */


  shouldRunRender() {
    const isTerminal = process.stdout.isTTY;

    if (this.command.report && !isTerminal) {
      return false;
    }

    return Boolean(this.command.render);
  }
  /**
   * this works for both, Harmony commands and Legacy commands (the legacy-command-adapter
   * implements json() method)
   */


  async runJsonHandler() {
    if (!this.flags.json) return null;
    if (!this.command.json) throw new Error(`command "${this.command.name}" doesn't implement "json" method`);
    const result = await this.command.json(this.args, this.flags);
    const code = result.code || 0;
    const data = result.data || result;
    return this.writeAndExit(JSON.stringify(data, null, 2), code);
  }

  async runRenderHandler() {
    if (!this.command.render) throw new Error('runRenderHandler expects command.render to be implemented');
    const result = await this.command.render(this.args, this.flags);

    _loader().default.off(); // @ts-ignore
    // eslint-disable-next-line no-prototype-builtins


    const data = result.data && result.hasOwnProperty('code') ? result.data : result; // @ts-ignore
    // eslint-disable-next-line no-prototype-builtins

    const code = result.data && result.hasOwnProperty('code') ? result.code : 0;
    const {
      waitUntilExit
    } = (0, _ink().render)(data);
    await waitUntilExit();
    return _logger().default.exitAfterFlush(code, this.command.name);
  }

  async runReportHandler() {
    if (!this.command.report) throw new Error('runReportHandler expects command.report to be implemented');
    const result = await this.command.report(this.args, this.flags);

    _loader().default.off();

    const data = typeof result === 'string' ? result : result.data;
    const exitCode = typeof result === 'string' ? 0 : result.code;
    return this.writeAndExit(`${data}\n`, exitCode);
  }
  /**
   * the loader and logger.console write output to the console during the command execution.
   * for internals commands, such as, _put, _fetch, the command.loader = false.
   */


  determineConsoleWritingDuringCommand() {
    if (this.command.loader && !this.flags.json) {
      _loader().default.on();

      _logger().default.shouldWriteToConsole = true;
    } else {
      _logger().default.shouldWriteToConsole = false;
    }

    if (this.flags.log) {
      const logValue = typeof this.flags.log === 'string' ? this.flags.log : undefined;

      _logger().default.switchToConsoleLogger(logValue);
    }
  }

  async writeAndExit(data, exitCode) {
    return process.stdout.write(data, async () => _logger().default.exitAfterFlush(exitCode, this.command.name));
  }

  async runMigrateIfNeeded() {
    // @ts-ignore LegacyCommandAdapter has .migration
    if (this.command.migration) {
      _logger().default.debug('Checking if a migration is needed'); // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!


      return (0, _consumer().migrate)(null, false);
    }

    return null;
  }

}

exports.CommandRunner = CommandRunner;

function serializeErrAndExit(err, commandName) {
  const data = (0, _utils().packCommand)((0, _utils().buildCommandMessage)((0, _serializeError().serializeError)(err), undefined, false), false, false);
  const code = err.code && (0, _utils().isNumeric)(err.code) ? err.code : 1;
  return process.stderr.write(data, () => _logger().default.exitAfterFlush(code, commandName));
}

async function handleErrorAndExit(err, commandName, shouldSerialize = false) {
  try {
    _loader().default.off();

    _logger().default.error(`got an error from command ${commandName}: ${err}`);

    _logger().default.error(err.stack || '<no error stack was found>');

    const {
      message,
      error
    } = (0, _defaultErrorHandler().default)(err);
    if (shouldSerialize) serializeErrAndExit(error, commandName);else await logErrAndExit(message, commandName);
  } catch (e) {
    // eslint-disable-next-line no-console
    console.error('failed to log the error properly, failure error', e); // eslint-disable-next-line no-console

    console.error('failed to log the error properly, original error', err);
    process.exit(1);
  }
}

async function handleUnhandledRejection(err) {
  // eslint-disable-next-line no-console
  console.error('** unhandled rejection found, please make sure the promise is resolved/rejected correctly! **');

  if (err instanceof Error) {
    return handleErrorAndExit(err, process.argv[2]);
  }

  console.error(err); // eslint-disable-line

  return handleErrorAndExit(new Error(`unhandledRejections found. err ${err}`), process.argv[2]);
}

async function logErrAndExit(err, commandName) {
  if (!err) throw new Error(`logErrAndExit expects to get either an Error or a string, got nothing`);
  console.error(err); // eslint-disable-line

  await _logger().default.exitAfterFlush(1, commandName);
}